//Optimización Unidimensional No Restringida
clc
//
//En este ejemplo demostraremos la optimización de una función utilizando los métodos de Newton,
//de interpolación parabólica sucesiva y de búsqueda dorada (Fibonacci modificado).
//El problema de optimización implica la búsqueda de un máximo o de un mínimo.
//En este ejemplo lo que buscamos es un máximo.
//
//Supongamos que tenemos la siguiente reacción batch secuencial:
//
//  A --> B --> C
//
//En este problema B es el producto deseado y C es un subproducto indeseado. Queremos
//maximizar el producto B. Las dos reacciones se suponen de primer orden:
//
//  Ra = -k1 * Ca
//
//  Rb = k1 * Ca - k2 * Cb
//
//donde k1 y k2 son constantes de reacción. Si la concentración inicial de B es cero  
//y la concentración inicial de A es Ca0 tenemos la solución:
//
//  Cb = Ca0*k1/(k2-k1)*(exp(-k1*t)-exp(-k2*t))
//
//que inicialmente aumenta y luego disminuye con el tiempo.
//Tenemos que definir esta función en un archivo de funciones de MATLAB, Cb.m.
//
//Grafiquemos esta función:
//En primer lugar elijamos valores de Ca0, k1 y k2 y hagámosle saber a la función
//Cb los valores que toman estos parámetros:
global Ca0 k1 k2
Ca0=1;
k1=2;
k2=1;
//
tt=[0:.1:2];
figure(1);
plot(tt,Cb(tt))
xlabel('Tiempo')
ylabel('Concentración de B')
title('Método de Newton')
gca().grid=[1 1 1]
set(gca(),"auto_clear","off")
//
//Como puede verse, existe un máximo alrededor de t = 0.7.
//Utilicemos las tres subrutinas de optimización discutidas en clase para determinar este valor
//con más exactitud.
//sleep(5,'s')
//
//En primer lugar, utilizamos el método de Newton arrancado de t=1.
//Tenemos la fórmula:
//
//  x(k+1) = x(k) - f'(x(k))/f''(x(k))
//
//Por simplicidad, calcularemos las derivadas numéricamente, utilizando algoritmos en
//diferencias finitas de segundo orden:
//
deltat=1e-5;
clear t
t(1)=1;
i=1;
Cbc=Cb(t(i));
plot(t(i),Cbc,'o')
fp=1;
dt=1;
while abs(dt)>1e-4
 Cbp=Cb(t(i)+deltat);
 Cbm=Cb(t(i)-deltat);
 fp=(Cbp-Cbm)/2/deltat;
 fpp=(Cbp+Cbm-2*Cbc)/deltat^2;
 dt=-fp/fpp;
 t(i+1)=t(i)+dt;
 i=i+1;
 Cbc=Cb(t(i));
 plot(t(i),Cbc,'o')
 sleep(1,'s')
end
disp('Método de Newton')
disp('Tiempo para el cual la concentración Cb es máxima: ')
disp(t(i))
disp('El valor máximo de la concentración Cb es: ')
disp(Cb(t(i)))
set(gca(),"auto_clear","on")
//
//La solución exacta para el máximo es:
texact=log(k2/k1)/(k2-k1) //Logaritmo natural
//
disp('la solución exacta para el máximo es: '),
disp(texact)
//Por lo tanto podemos calcular la velocidad de progresión a la solución correcta:
//
ernm=abs(t-texact);
nnm=length(ernm);
figure(2);
plot2d("ll",ernm(1:nnm-1),ernm(2:nnm))
xlabel('e(i)')
ylabel('e(i+1)')
title('Progresión del error en el método de Newton')
gca().grid=[1 1 1]
set(gca(),"auto_clear","on")
//Consecuentemente podemos ver que para esta función el método de Newton converge
//cuadráticamente. Obsérvese que esto funciona mientras podamos determinar la primera y segunda
//derivada con suficiente seguridad.
//
//También se requiere comenzar donde la segunda derivada es del signo apropiado, en este caso
//negativo. Si hubiésemos comenzado en t=2 habríamos capturado el mínimo para t --> infinito.
//Veamos las gráficas de las derivadas primera y segunda de la concentración de B:
//
figure(3);
Cb12=[Cb1(tt)',Cb2(tt)'];
plot2d('nn',tt',[Cb12(:,1),Cb12(:,2)],[1:2],leg='Derivada Primera de la Concentración de B@Derivada Segunda de la Concentración de B')
xlabel('Tiempo')
ylabel('Función')
title('Derivadas de la Concentración')
gca().grid=[1 1 1]
set(gca(),"auto_clear","off")
//
//Veamos ahora el método de interpolación parabólica sucesiva. En cada etapa ajustamos la
//parábola y(t)=x(1)+x(2)*t+x(3)*t^2 a tres puntos y la utilizamos para estimar donde está el máximo o el mínimo.
//Comenzamos:
//
clear t
t(1)=1;
t(2)=0;
t(3)=.5;
f=Cb(t);
i=3;
figure(4);
plot(tt,Cb(tt))
xlabel('Tiempo')
ylabel('Concentración de B')
title('Método de Interpolación Parabólica Sucesiva')
gca().grid=[1 1 1]
set(gca(),"auto_clear","off")
plot(t,f,'o')
//
while abs(t(i)-t(i-1))>1e-7,
 tm=t(i-2:i);
 fm=f(i-2:i);
 a=[ones(3,1),tm,tm.^2];
 x=a\fm;
 i=i+1;
 t(i)=-x(2)/2/x(3);
 f(i)=Cb(t(i));
 plot(t(i),f(i),'o')
sleep(1,'s')
end
disp('Método de Interpolación Parabólica Sucesiva')
disp('Tiempo para el cual la concentración Cb es máxima: ')
disp(t(i))
disp('El valor máximo de la concentración Cb es: ')
disp(Cb(t(i)))
set(gca(),"auto_clear","on")
//
//Podemos calcular la velocidad de progresión a la solución exacta:
//
erspi=abs(t-texact);
nspi=length(erspi);
figure(5);
//plot2d("ll",ernm(1:nnm-1),ernm(2:nnm),leg='Método de Newton')
set(gca(),"auto_clear","off")
plot2d("ll",erspi(1:nspi-1),erspi(2:nspi))
gca().grid=[1 1 1]
xlabel('e(i)')
ylabel('e(i+1)')
title('Progresión del Error en el Método de Interpolación Parabólica Sucesiva')
set(gca(),"auto_clear","on")
//
//En este caso la velocidad de progresión es ligeramente super lineal, con una pendiente
//de 1.324. Esta progresión también es un poco más aleatoria.
//
//Finalizamos con el método de Fibonacci o de la búsqueda dorada. Este método utiliza la
//relación o razón dorada que es solución de:
//
//  r^2 = 1 - r
//
//o r = 0.6180. Esta búsqueda sólo funciona sobre funciones unimodales, esto es, aquellas
//funciones para las cuales la primera derivada es estrictamente positiva para t < t* y negativa
//para t > t*, o a la inversa si estamos buscando un mínimo. Obsérvese que la segunda derivada
//puede cambiar de signo localmente y el método ser aún válido.
//El método de Newton debería fallar en esta circunstancia.
//
//Trabajemos este método. Comenzamos con algún intervalo conocido sobre el cual la función
//es unimodal, digamos [0,1]. Evaluamos la función en el interior en dos puntos adicionales
//localizados en r y 1-r, en la parte derecha e izquierda, respectivamente, del
//intervalo de búsqueda [0,1]. 
//En el caso más general que el intervalo de búsqueda sea [a,b], el punto interior izquierdo
//es xi = a+(b-a)*(1-r), mientras que el punto interior derecho es xr = a+(b-a)*r.
//Conservamos el nuevo intervalo [0,r] o [1-r,1]([a,xr] o [xi,b]) dependiendo
//cuál contiene el máximo (o el mínimo). Entonces procedemos a la próxima etapa. 
//La clave es que el método sólo requiere una nueva evaluación de la función en cada etapa.
sleep(15,'s')
//Hagamos esto:
//
r=(sqrt(5)-1)/2; //r=0.618
clear t
figure(6);
plot(tt,Cb(tt))
xlabel('Tiempo')
ylabel('Concentración de B')
title('Método de Fibonacci')
gca().grid=[1 1 1]
set(gca(),"auto_clear","off")
a=0;
b=1;
t(1)=a+(b-a)*(1-r);//Representación general del tiempo en el lado izquierdo del intervalo [a,b]
tl=t(1);//tl: t en el lado izquierdo del intervalo o t left
t(2)=a+(b-a)*r;//Representación general del tiempo en el lado derecho del intervalo [a,b]
tr=t(2);//tr: t en el lado derecho o t right
f=Cb(t);
plot(t,f,'o')
fl=f(1);
fr=f(2);
i=2;
//
while abs(t(i)-t(i-1))>1e-5,
 i=i+1;
 if fr>fl,
  a=tl;//Se descarta el lado izquierdo del punto interior izquierdo
//El punto interior izquierdo del nuevo intervalo es igual al punto
//interior derecho del viejo intervalo
  tl=tr;
  fl=fr;
//Se calcula el nuevo punto interior derecho
  t(i)=a+(b-a)*r;
  tr=t(i);
  fr=Cb(tr);
  plot(tr,fr,'o')
 else,
  b=tr;//Se descarta el lado derecho del punto interior derecho
 //El punto interior derecho del nuevo intervalo es igual al punto
 //interior izquierdo del viejo intervalo 
  tr=tl;
  fr=fl;
 //Se calcula el nuevo punto interior derecho
  t(i)=a+(b-a)*(1-r);
  tl=t(i);
  fl=Cb(tl);
  plot(tl,fl,'o')
 end
 sleep(1,'s')
end
disp('Método de Fibonacci')
disp('Tiempo para el cual la concentración Cb es máxima: ')
disp(t(i))
disp('El valor máximo de la concentración Cb es: ')
disp(Cb(t(i)))
set(gca(),"auto_clear","on")
//
//Esta técnica converge aún más lentamente que el método de interpolación parabólica sucesiva.
//La velocidad de convergencia es lineal y el intervalo se reduce en un factor 0.618 en cada
//etapa. Podemos graficar esto: 
//
ergs=abs(t-texact);
ngs=length(ergs);
figure(7);
plot2d("ll",ernm(1:nnm-1),ernm(2:nnm))
xlabel('e(i)')
ylabel('e(i+1)')
title('Progresión de los Errores')
gca().grid=[1 1 1]
set(gca(),"auto_clear","off")
plot2d("ll",erspi(1:nspi-1),erspi(2:nspi))
plot2d("ll",ergs(1:ngs-1),ergs(2:ngs))
//
//Si comparamos estas tres técnicas para esta función vemos que el método de Newton es el que
//converge más rápido pero sólo si arrancamos de un valor próximo al valor verdadero. El método
//de la búsqueda dorada es más lento pero más robusto -podríamos haber elegido cualquier
//intervalo conteniendo al máximo (mínimo) y eventualmente haber llegado al valor correcto-. No
//se puede decir lo mismo de las dos técnicas superlineales.
